import datetime
import asyncio
from collections import namedtuple

ImageResult = namedtuple("ImageResult", ["tags", "source", "rating"])
"""The response from the engine.

.. py:attribute:: tags

    The tags the engine has located.

.. py:attribute:: source

    The source that has been found for the image.

.. py:attribute:: rating

    The rating on the image.

"""


class Engine:
    """Base class for an engine.
    
    All the engine classes must derive from this class.

    :ivar host: The base URL for the reverse image domain. This is not an API endpoint, but can be a link to IQDB or something similar.
    :vartype host: str
    :ivar download_required: Determines if the ``search_image()`` function will download the image to be searched beforehand or not.
    :vartype download_required: bool
    :ivar datetime.datetime ~.last_request: Optional. Instance variable that can be set to make use of :meth:`Booru.sleep_until_ratelimit`. Update after making a request. If you just use the library, this should be done by the engine so there's no need to manually set it.
    """
    host = None
    download_required = False

    def __init__(self): # pragma: no cover
        self.last_request = None
        raise NotImplementedError("Expand this method to include all needed keys.")

    async def search_tag(self, tag): # pragma: no cover
        """Reverse search the booru for tag data.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")

    async def search_image(self, url): # pragma: no cover
        """Reverse search the engine for ``url``.

        :param str url: URL to search
        :rtype: ImageResult
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")

    async def sleep_until_ratelimit(self, sleep_time): # pragma: no cover
        """This is an additional helper function for APIs with known rate limits.

        If an engine can hit the ratelimit and it's timeout is acceptable (an acceptable timeframe is no more than one minute),
        this helper function will allow the engine to sleep until it has passed.

        If the timeframe is not acceptable, raise :class:`EngineCooldownException` instead.

        To use this, you will need to add the instance variable last_request See the main class doc for details.
        """
        timeout = datetime.timedelta(seconds=sleep_time) # 35 seconds -> ensures no issue
        until = self.last_request + timeout
        now = datetime.datetime.now()
        if now > until:
            return # Fine, limit has passed
        else:
            await asyncio.sleep((datetime.datetime.now() - until).seconds)


import datetime
import asyncio
import functools

from retaggr.engines.base import Engine, ImageResult
from retaggr.errors import NotAvailableSearchException, EngineCooldownException
import requests as fuck_aiohttp

class SauceNao(Engine):
    """Reverse searches the SauceNao API and then does additional matching.

    This booru does not require images to be downloaded before searching.

    This API is subject to rate limits.

    :param api_key: SauceNao API key. You can get this by registering an account on saucenao.com
    :type api_key: str
    """
    host = "https://saucenao.com"
    download_required = False


    tag_indexes = set([9, 12, 26])
    """Tag indexes we can parse.

    Valid index numbers can be found at https://saucenao.com/status.html .

    List of indexes we can parse for tags:

    * 9: Danbooru
    * 12: Yande.re
    * 26: Konachan
    """

    source_indexes = [5, 16, 37, 34]
    """List of source indexes in preferred order (key 0 is preferred, last key is least preferred).
    
    Valid index numbers can be found at https://saucenao.com/status.html .

    List of indexes we use for sources:

    * 5: Pixiv (preferred, low quantity/risk of reuploads)
    * 16: FAKKU (official redistribution)
    * 37: MangaDex (not official redistribution, but metadata is accurate)
    * 34: DeviantART (not preferred, large number of art theft and reuploads)
    """

    long_remaining = 2
    """The total amount of requests remaining for the next 24 hours. If this is 0, the library will raise an :class:`retaggr.errors.EngineCooldownException` until the rate limit expires."""

    short_remaining = 2
    """The total amount of requests remaining for the next 30 seconds. If this is 0, the library will sleep until it expires."""

    last_request = None
    """The last request date the engine made to saucenao. Used to regulate the ratelimits."""

    def __init__(self, api_key):
        self.api_key = api_key
        self.last_request = datetime.datetime.now()

    async def search_image(self, url):
        request_url = "https://saucenao.com/search.php"
        params = {
            "db": "999", # No clever bitmasking -> need help with how to do that.
            "api_key": self.api_key,
            "output_type": "2", # 2 is the JSON API,
            "url": url
        }
        if self.last_request: # pragma: no cover
            long_ratelimit_ends = self.last_request + datetime.timedelta(hours=24)
            if self.long_remaining <= 1 and long_ratelimit_ends < datetime.datetime.now():
                raise EngineCooldownException(until=long_ratelimit_ends)

        if self.short_remaining <= 1: # pragma: no cover
            await self.sleep_until_ratelimit(35) # 35 seconds = no issue

        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, request_url, params=params))
        j = r.json()
        self.last_request = datetime.datetime.now()
        if r.status_code == 200:
            self.short_remaining = j["header"]["short_remaining"] - 1
            self.long_remaining = j["header"]["long_remaining"] - 1
            return await self.index_parser(j)
        elif r.status_code == 429: # 429 indicates we hit a rate limit, but not one we could've accounted for before.
            self.short_remaining = 0
            # We do raise an EngineCooldownException, since this indicates that it occured before API instantiating a request
            # (rest of library handles it with a previously done request already)
            raise EngineCooldownException(until=datetime.datetime.now() + datetime.timedelta(seconds=35)) 

    async def search_tag(self, tag):
        raise NotAvailableSearchException("This engine cannot search tags.")

    async def index_parser(self, json):
        """Parse the output from a succesful saucenao search to retrieve data from specific indexes.
        
        :param json: JSON output from the API.
        :type json: dict
        :return: Dictionary containing data that matches the output for :meth:`SauceNao.search_image_source`
        :rtype: dict
        """
        base_similarity = json["header"]["minimum_similarity"] # Grab the minimum similarity saucenao advises, going lower is generally gonna give false positives.

        # Below we cast the _entry_ similarity to a float since somehow it's stored as an str.
        # Damn API inaccuracy
        source_results = [entry for entry in json["results"] if entry["header"]["index_id"] in self.source_indexes and float(entry["header"]["similarity"]) > base_similarity]

        source = None
        source_priority = len(self.source_indexes) # No result priority is 1 above the least wanted result
        for entry in source_results:
            list_index = self.source_indexes.index(entry["header"]["index_id"])
            if list_index < source_priority: # If our exsting result priority is lower than the current result...
                source = entry["data"]["ext_urls"][0] # We update the source with that result
                source_priority = list_index # And we update the priority

        # See my comment above source_results as to what I'm doing here.
        tag_results_list = [entry for entry in json["results"] if entry["header"]["index_id"] in self.tag_indexes and float(entry["header"]["similarity"]) > base_similarity]

        # Unlike the source, these require specific lookups based on their ID. As a result, I'll rearrange the results to a dict.
        tag_results = {}
        for entry in tag_results_list:
            tag_results[entry["header"]["index_id"]] = entry

        # Kinda looks stupid, but whatever.
        loop = asyncio.get_event_loop()
        tags = []
        for index_id, entry in tag_results.items():
            if index_id == 9: # Danbooru
                r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://danbooru.donmai.us/posts/" + str(entry["data"]["danbooru_id"]) + ".json"))
                j = r.json()
                tags += j["tag_string"].split()
            if index_id == 12: # Yande.re # pragma: no cover
                r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://yande.re/post.json", params={"tags": "id:" + str(entry["data"]["yandere_id"])}))
                j = r.json()
                tags += j[0]["tags"].split()
            if index_id == 26: # Konachan # pragma: no cover
                r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "http://konachan.com/post.json", params={"tags": "id:" + str(entry["data"]["konachan_id"])}))
                j = r.json()
                tags += j[0]["tags"].split()

        return ImageResult(tags, source, None)


# stdlib
from collections import namedtuple

# Config
from retaggr.config import ReverseSearchConfig

# Boorus
from retaggr.engines.danbooru import Danbooru
from retaggr.engines.e621 import E621
from retaggr.engines.iqdb import Iqdb
from retaggr.engines.paheal import Paheal
from retaggr.engines.saucenao import SauceNao

# Exceptions
from retaggr.errors import MissingAPIKeysException, NotAValidEngineException, EngineCooldownException

ReverseResult = namedtuple("ReverseResult", ["tags", "source", "rating"])
"""The response from a reverse image search. All attributes are Sets.

.. py:attribute:: tags

    The tags the engine has located.

.. py:attribute:: source

    The source that has been found for the image.

.. py:attribute:: rating

    The rating on the image.

"""

class ReverseSearch:
    r"""Core class used for Reverse Searching. 
    
    This class can only be instantiated with a :class:`ReverseSearchConfig` instance.

    All listed methods can only be ran from an asynchronous context.

    :ivar accessible_engines: The accessible boorus from the passed in configuration object.
    :param config: The config object.
    :type config: ReverseSearchConfig
    """
    _all_engines = [
        "danbooru",
        "e621",
        "iqdb",
        "paheal",
        "saucenao"
        ]

    def __init__(self, config):
        self.config = config
        self.accessible_engines = {}

        if hasattr(self.config, "min_score"):
            if hasattr(self.config, "danbooru_username") and hasattr(self.config, "danbooru_api_key"):
                self.accessible_engines["danbooru"] = Danbooru(self.config.danbooru_username, self.config.danbooru_api_key, self.config.min_score)
            if hasattr(self.config, "e621_username") and hasattr(self.config, "app_name") and hasattr(self.config, "version"):
                self.accessible_engines["e621"] = E621(self.config.e621_username, self.config.app_name, self.config.version, self.config.min_score)
            self.accessible_engines["iqdb"] = Iqdb(self.config.min_score)

        if hasattr(self.config, "saucenao_api_key"):
            self.accessible_engines["saucenao"] = SauceNao(self.config.saucenao_api_key)

        self.accessible_engines["paheal"] = Paheal()

    async def reverse_search(self, url, callback=None, download=False):
        """
        Reverse searches all accessible boorus for ``url``.

        .. note::
            Callback is a callback function that can be passed in. This can be used to keep track of
            progress for certain methods and functions.

            .. code-block:: python
               :linenos:

                async def callback(engine, rresult):
                    print("This booru was searched: %s", booru)
                    print("These tags were found: %s", rresult.tags)
                    print("This source was found: %s", rresult.source)
                    print("This rating was found: %s", rresult.rating)

                # Callback will be called each time a search finishes.
                rs.reverse_search(url, callback)

        After a reverse search, this method will be called with the engine name that was just searched.

        :param url: The URL to search.
        :type url: str
        :param callback: Callback function.
        :type callback: Optional[function]
        :param download: Run searches on boorus that require a file download. Defaults to False.
        :type download: Optional[bool]
        :return: A :class:`ReverseResult` instance containing your data.
        :rtype: ReverseResult
        """
        tags = set()
        source = set()
        rating = set()
        for engine in self.accessible_engines:
            if self.accessible_engines[engine].download_required:
                if not download:
                    continue
            try:
                result = await self.search_image(engine, url)
            except EngineCooldownException: # pragma: no cover
                pass
            else:
                tags.update(result.tags)
                if result.source:
                    source.update(result.source)
                if result.rating:
                    rating.update(result.rating)
                if callback:
                    await callback(engine, ReverseResult(result.tags, result.source, result.rating))
        return ReverseResult(tags, source, rating)

    async def search_image(self, booru, url):
        r"""Reverse search a booru for ``url``.

        :param booru: Booru to search, this must match a filename in the boorus folder.
        :type booru: str
        :param url: The URL to search.
        :type url: str
        :raises MissingAPIKeysException: Required keys in config object missing.
        :raises NotAValidEngineException: The passed in booru is not a valid booru.
        :return: A :class:`ImageResult` instance containing your data.
        :rtype: ImageResult
        """
        if booru not in self._all_engines:
            raise NotAValidEngineException("%s is not a valid engine", booru)
        if booru not in self.accessible_engines:
            raise MissingAPIKeysException("%s is misisng one or more needed API keys. Check the documentation.")
        return await self.accessible_engines[booru].search_image(url)



from retaggr.engines.base import Engine, ImageResult

# External modules
import asyncio
import requests as fuck_aiohttp
from aiohttp_requests import requests
import functools
import time
from lxml import html
import datetime

class E621(Engine):
    """Reverse searches https://e621.net through using the IQDB implementation over at https://iqdb.harry.lu in combination
    with the official e621 API.
    
    This booru does not required images to be downloaded before searching.

    :param username: Your e621 username
    :type username: str
    :param app_name: The name of your application.
    :type app_name: str
    :param version: The version of your application.
    :type version: float
    :param min_score: Minimum search match percentage needed.
    :type min_score: float
    """

    host = "https://iqdb.harry.lu/"
    e621_api = "https://e621.net/post/show.json"
    download_required = False

    last_request = None
    """The last request date the engine made to E621. Used to regulate the ratelimits."""


    def __init__(self, username, app_name, version, min_score):
        self.user_agent = {"User-Agent": f"{app_name}/{version} (by {username} on e621)"}
        self.min_score = min_score
        self.last_request = datetime.datetime.now()

    async def search_image(self, url):
        results = {
            "tags": [],
            "source": None
        }

        params = {"url" : url}

        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.post, self.host, params=params))

        doc = html.fromstring(r.text)
        tables = doc.xpath("//div[@id='pages']/div/table/tr/td")
    
        row = 5
        while row < len(tables):
            try:
                percent = float(str.split(tables[row].text)[0][:-1])
                # Get tags from page
                if percent > self.min_score:
                    post_id = tables[row - 3].xpath("//td/a")[0].get("href").split("/")[-1]
                    # Check tags from api
                    await self.sleep_until_ratelimit(1)
                    r = await requests.get(self.e621_api, headers=self.user_agent, params={"id": post_id})
                    self.last_request = datetime.datetime.now()
                    json = await r.json()
                    tags = json['tags'].split()
                    source = json['source']
                    rating = json['rating']
            except: # pragma: no cover
                pass
            row = row + 5
        return ImageResult(tags, source, rating)


class MissingAPIKeysException(Exception):
    """Raised if a required API key to search an engine is missing."""
    pass

class NotAValidEngineException(Exception):
    """Raised if the passed in engine does not exist."""

class NotAvailableSearchException(Exception):
    """This engine is not capable of searching this option."""

class EngineCooldownException(Exception):
    """This engine is on cooldown.
    
    :ivar datetime.datetime until: The period for which this engine is on cooldown."""

    def __init__(self, until):
        self.until = until